/* INCLUDE FILES */
#include "general.h"    /* always include first */
#include "parse.h"      /* always include */
#include "debug.h"
#include "get.h"
#include "options.h"
#include "read.h"
#include "routines.h"

#include <setjmp.h>
#include <stdio.h>
#include <string.h>
	
/* so we can call the C++ parser to do the non-ObjectiveC work */
extern boolean findCTags (const unsigned int passCount);

/* DATA DEFINITIONS */
static jmp_buf Exception;

static boolean gParsingInterface;
static boolean gParsingImplementation;

typedef enum eObjectiveCKinds {
	K_CLASS, K_FUNCTION, K_PROTOTYPE
} objCKind;

static kindOption ObjCKinds [] = {
	{ TRUE, 'c', "class",    "classes" },
	{ TRUE, 'f', "function", "functions" },
	{ TRUE, 'p', "prototype",  "function prototypes"},
};

typedef enum eTokenType {
	TOKEN_NONE,          /* none */
	TOKEN_ARGS,          /* a parenthetical pair and its contents */
	TOKEN_BRACE_CLOSE,
	TOKEN_BRACE_OPEN,
	TOKEN_COLON,         /* the colon character */
	TOKEN_COMMA,         /* the comma character */
	TOKEN_KEYWORD,
	TOKEN_NAME,          /* an unknown name */
	TOKEN_PAREN_NAME,    /* a single name in parentheses */
	TOKEN_SEMICOLON,     /* the semicolon character */
	TOKEN_SPEC,          /* a storage class specifier, qualifier, type, etc. */
	TOKEN_COUNT
} tokenType;

typedef struct sTokenInfo {
	tokenType     type;
	vString*      name;          /* the name of the token */
	unsigned long lineNumber;    /* line number of tag */
	fpos_t        filePosition;  /* file position of line containing name */
} tokenInfo;

typedef enum eException {
	ExceptionNone, ExceptionEOF, ExceptionFormattingError,
	ExceptionBraceFormattingError
} exception_t;

typedef enum eTagType {
	TAG_CLASS,       /* class name */
	TAG_FUNCTION,    /* function definition */
	TAG_PROTOTYPE,   /* structure, class or interface member */
	TAG_COUNT        /* must be last */
} tagType;


/* FUNCTION DEFINITIONS */

static void makeTag (const tokenInfo *const token, const tagType type)
{
	if (vStringLength (token->name) > 0 )
	{
		vString *scope = vStringNew ();
		/* Use "typeRef" to store the typename from addOtherFields() until
		 * it's used in makeTagEntry().
		 */
		vString *typeRef = vStringNew ();
		tagEntryInfo e;

		initTagEntry (&e, vStringValue (token->name));

		e.lineNumber	= token->lineNumber;
		e.filePosition	= token->filePosition;
		e.isFileScope	= ! isHeaderFile ();
		e.kindName		= ObjCKinds [type].name;
		e.kind			= ObjCKinds [type].letter;

		makeTagEntry (&e);
		vStringDelete (scope);
		vStringDelete (typeRef);
	}
}

static void initToken (tokenInfo* const token)
{
	token->type			= TOKEN_NONE;
	token->lineNumber	= getSourceLineNumber ();
	token->filePosition	= getInputFilePosition ();
	vStringClear (token->name);
}

static tokenInfo *newToken (void)
{
	tokenInfo *const token = xMalloc (1, tokenInfo);
	token->name = vStringNew ();
	initToken (token);
	return token;
}

static void deleteToken (tokenInfo *const token)
{
	if (token != NULL)
	{
		vStringDelete (token->name);
		eFree (token);
	}
}

static int skipToOneOf (const char *const chars)
{
	int c;
	do
		c = cppGetc ();
	while (c != EOF  &&  c != '\0'  &&  strchr (chars, c) == NULL);
	return c;
}

/*  Skip to the next non-white character.
 */
static int skipToNonWhite (void)
{
	boolean found = FALSE;
	int c;

	while (1)
	{
		c = cppGetc ();
		if (isspace (c))
			found = TRUE;
		else
			break;
	}

	return c;
}

/*  Skips to the next brace in column 1. This is intended for cases where
 *  preprocessor constructs result in unbalanced braces.
 */
static void skipToFormattedBraceMatch (void)
{
	int c, next;

	c = cppGetc ();
	next = cppGetc ();
	while (c != EOF  &&  (c != '\n'  ||  next != '}'))
	{
		c = next;
		next = cppGetc ();
	}
}

/*  Skip to the matching character indicated by the pair string. If skipping
 *  to a matching brace and any brace is found within a different level of a
 *  #if conditional statement while brace formatting is in effect, we skip to
 *  the brace matched by its formatting. It is assumed that we have already
 *  read the character which starts the group (i.e. the first character of
 *  "pair").
 */
static void skipToMatch (const char *const pair)
{
	const boolean braceMatching = (boolean) (strcmp ("{}", pair) == 0);
	const boolean braceFormatting = (boolean) (isBraceFormat () && braceMatching);
	const unsigned int initialLevel = getDirectiveNestLevel ();
	const int begin = pair [0], end = pair [1];
	const unsigned long inputLineNumber = getInputLineNumber ();
	int matchLevel = 1;
	int c = '\0';

	while (matchLevel > 0  &&  (c = skipToNonWhite ()) != EOF)
	{
		if (c == begin)
		{
			++matchLevel;
			if (braceFormatting  &&  getDirectiveNestLevel () != initialLevel)
			{
				skipToFormattedBraceMatch ();
				break;
			}
		}
		else if (c == end)
		{
			--matchLevel;
			if (braceFormatting  &&  getDirectiveNestLevel () != initialLevel)
			{
				skipToFormattedBraceMatch ();
				break;
			}
		}
	}
	if (c == EOF)
	{
		verbose ("%s: failed to find match for '%c' at line %lu\n",
				getInputFileName (), begin, inputLineNumber);
		if (braceMatching)
			longjmp (Exception, (int) ExceptionBraceFormattingError);
		else
			longjmp (Exception, (int) ExceptionFormattingError);
	}
}


static void skipParens (void)
{
	int c = skipToNonWhite ();

	if (c == '(')
	{
		skipToMatch ("()");
		c = skipToNonWhite ();
	}

	cppUngetc (c);
}

static void skipBraces (void)
{
	const int c = skipToNonWhite ();

	if (c == '{')
		skipToMatch ("{}");
	else
		cppUngetc (c);
}

static void readWordIntoToken (tokenInfo *const token, const int firstChar)
{
	vString *const name = token->name;
	int c = firstChar;

	do
	{
		vStringPut (name, c);
		c = cppGetc ();
	} while (isalnum(c));
	
	vStringTerminate (name);
	cppUngetc (c);        /* unget character that dropped us out of the loop */
}

static void reformatMethodTokenName(tokenInfo *token)
{
	/* there are three flavors of method we need to parse out:
	    no parameters:
			+sharedWebDelegate
		one parameter:
			-markInfoForUseOnWebViewReload:(WebView *)webview
		more than one parameter:
			-rememberVisibleRect:(NSRect)visibleRect contentSize:(NSSize)contentSize forWebView:(WebView *)webView
		
		Ultimately, we want all the words immediately to the left of a colon, so we'll start at the end of the string, and work backward...
	 */
	 
	vString *const name = token->name;
	boolean foundOne = FALSE;

	char *newStringMemory = (char*)malloc(name->size); /* it can only get shorter */
//	memset(newString, -1, name->size);
	bzero(newStringMemory, name->size);
	char * newString = newStringMemory + name->size-2;
	
	ssize_t oldIndex = (ssize_t)(name->size -1);
	
	while (0 <= oldIndex)
	{
		/* look for the first ':' */
		if (':' == name->buffer[oldIndex])
		{
			foundOne = TRUE;
			do
			{
				*(--newString) = name->buffer[oldIndex];
				oldIndex--;
			} while (' ' != name->buffer[oldIndex] && (0 <= oldIndex));
		}

		oldIndex--;
	}
	
	if (foundOne)
	{
		vStringClear (name);
		vStringCopyS (name, newString);
	}

	vStringStripNewline(name);
	
	free(newStringMemory);
}

static void readMethodIntoToken(tokenInfo *token, const int firstChar)
{
	if (gParsingInterface || gParsingImplementation)
	{
		vString *const name = token->name;
		int c = firstChar;
		
		do
		{
			vStringPut (name, c);
			c = cppGetc ();
		} while (c != EOF && c != '\0' && strchr (";{", c) == NULL);
		
		vStringTerminate (name);
		cppUngetc (c);        /* unget character that dropped us out of the loop */
		skipBraces ();
	}
}

static void setStateForToken(tokenInfo *token)
{
	size_t tokenLen = token->name->length;
	
	if (tokenLen == strlen("@implementation") && (0 == strncmp(token->name->buffer, "@implementation", tokenLen)) )
	{
		Assert(! gParsingInterface);
		Assert(! gParsingImplementation);

		gParsingImplementation = TRUE;
	}
	else if (tokenLen == strlen("@interface") && (0 == strncmp(token->name->buffer, "@interface", tokenLen)) )
	{
		Assert(! gParsingInterface);
		Assert(! gParsingImplementation);

		gParsingInterface = TRUE;
	}
	else if (tokenLen == strlen("@end") && (0 == strncmp(token->name->buffer, "@end", tokenLen)) )
	{
		Assert(gParsingInterface != gParsingImplementation);

		gParsingInterface = FALSE;
		gParsingImplementation = FALSE;
	}
	/* else it's a string, selector, or something else that doesn't alter our state */
}

static void findObjCTags(void)
{	
	/* start off by getting all the C/C++ data from the file */
	findCTags (1);
	
	/* reset the file pointer and internal bookkeeping */
	fileRewind ();
		
	cppInit(TRUE, FALSE);
	
	/* then, start digging out the Objective C stuff */
	do
	{
		int c = skipToOneOf ("@+-");
		if (-1 == c)
			break;
		
		tokenInfo *token = newToken ();
		
		if ('@' == c)
		{
			readWordIntoToken (token, c);
			setStateForToken (token);
			
			if (gParsingInterface)
			{
				deleteToken (token);
				token = newToken ();
				c = skipToNonWhite ();
				readWordIntoToken (token, c);
				makeTag (token, TAG_CLASS);
			}
		}
		else if (gParsingInterface || gParsingImplementation)
		{
			/* eat optional whitespace, and the parenthesized return value -- read the rest into the token */
			skipParens ();

#define skipPlusMinus 1
#ifdef skipPlusMinus
			c = skipToNonWhite ();
#endif
			readMethodIntoToken (token, c);
			reformatMethodTokenName (token);
			makeTag (token, gParsingInterface ? TAG_PROTOTYPE : TAG_FUNCTION);
		}
		deleteToken (token);

	} while (1);
	
	/* final cleanup */
	cppTerminate ();
}

/* Create parser definition stucture */
extern parserDefinition* ObjectiveCParser(void)
{
    static const char *const extensions [] = { "m", "mm", NULL };
    parserDefinition* def = parserNew ("Objective-C");
    def->kinds      = ObjCKinds;
    def->kindCount  = KIND_COUNT(ObjCKinds);
    def->extensions = extensions;
    def->parser     = findObjCTags;
    return def;
}
